using System;
using System.Reflection;
using System.Collections.Generic;
using System.Linq;
using Encompass.Exceptions;

namespace Encompass
{
    public abstract class Engine
    {
        internal readonly HashSet<Type> sendTypes = new HashSet<Type>();
        internal readonly HashSet<Type> receiveTypes = new HashSet<Type>();

        private EntityManager entityManager;
        private ComponentManager componentManager;
        private MessageManager messageManager;
        private ComponentMessageManager componentMessageManager;

        protected Engine()
        {
            var sendsAttribute = GetType().GetCustomAttribute<Sends>(false);
            if (sendsAttribute != null)
            {
                sendTypes = sendsAttribute.sendTypes;
            }

            var activatesAttribute = GetType().GetCustomAttribute<Activates>(false);
            if (activatesAttribute != null)
            {
                sendTypes.UnionWith(activatesAttribute.activateTypes);
            }

            var updatesAttribute = GetType().GetCustomAttribute<Updates>(false);
            if (updatesAttribute != null)
            {
                sendTypes.UnionWith(updatesAttribute.updateTypes);
            }

            var receivesAttribute = GetType().GetCustomAttribute<Receives>(false);
            if (receivesAttribute != null)
            {
                receiveTypes = receivesAttribute.receiveTypes;
            }

            var readsAttribute = GetType().GetCustomAttribute<Reads>(false);
            if (readsAttribute != null)
            {
                receiveTypes.UnionWith(readsAttribute.readTypes);
            }

            var readsPendingAttribute = GetType().GetCustomAttribute<ReadsPending>(false);
            if (readsPendingAttribute != null)
            {
                receiveTypes.UnionWith(readsPendingAttribute.readPendingTypes);
            }
        }

        internal void AssignEntityManager(EntityManager entityManager)
        {
            this.entityManager = entityManager;
        }

        internal void AssignComponentManager(ComponentManager componentManager)
        {
            this.componentManager = componentManager;
        }

        internal void AssignMessageManager(MessageManager messageManager)
        {
            this.messageManager = messageManager;
        }

        internal void AssignComponentMessageManager(ComponentMessageManager componentMessageManager)
        {
            this.componentMessageManager = componentMessageManager;
        }

        public abstract void Update(double dt);

        protected Entity CreateEntity()
        {
            return entityManager.CreateEntity();
        }

        protected bool EntityExists(Guid entityID)
        {
            return entityManager.EntityExists(entityID);
        }

        protected Entity GetEntity(Guid entityID)
        {
            return entityManager.GetEntity(entityID);
        }

        protected Guid GetEntityIDByComponentID(Guid componentID)
        {
            return componentManager.GetEntityIDByComponentID(componentID);
        }

        protected Entity GetEntityByComponentID(Guid componentID)
        {
            return GetEntity(GetEntityIDByComponentID(componentID));
        }

        protected TComponent GetComponentByID<TComponent>(Guid componentID) where TComponent : struct, IComponent
        {
            if (componentManager.GetComponentTypeByID(componentID) != typeof(TComponent))
            {
                throw new ComponentTypeMismatchException("Expected Component to be of type {0} but was actually of type {1}", typeof(TComponent).Name, componentManager.GetComponentTypeByID(componentID).Name);
            }

            return (TComponent)componentManager.GetComponentByID(componentID);
        }

        internal IEnumerable<(Guid, TComponent)> ReadComponentsFromWorld<TComponent>() where TComponent : struct, IComponent
        {
            return componentManager.GetComponentsByType<TComponent>();
        }

        protected Guid AddComponent<TComponent>(Entity entity, TComponent component) where TComponent : struct, IComponent
        {
            var componentID = componentManager.NextID();

            componentManager.AddComponent(entity, componentID, component);

            if (sendTypes.Contains(typeof(PendingComponentMessage<TComponent>)))
            {
                PendingComponentMessage<TComponent> newComponentMessage;
                newComponentMessage.entity = entity;
                newComponentMessage.componentID = componentID;
                newComponentMessage.component = component;
                SendMessage(newComponentMessage);
                SendPendingComponentMessage(newComponentMessage);
            }

            return componentID;
        }

        protected Guid AddDrawComponent<TComponent>(Entity entity, TComponent component, int layer = 0) where TComponent : struct, IComponent
        {
            var componentID = componentManager.NextID();

            componentManager.AddDrawComponent(entity, componentID, component, layer);

            if (sendTypes.Contains(typeof(PendingComponentMessage<TComponent>)))
            {
                PendingComponentMessage<TComponent> newComponentMessage;
                newComponentMessage.entity = entity;
                newComponentMessage.componentID = componentID;
                newComponentMessage.component = component;
                SendMessage(newComponentMessage);
                SendPendingComponentMessage(newComponentMessage);
            }

            return componentID;
        }

        protected IEnumerable<ValueTuple<Guid, TComponent>> ReadComponents<TComponent>() where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));
            if (existingRead && pendingRead)
            {
                return componentMessageManager.ReadExistingAndPendingComponentsByType<TComponent>();
            }
            else if (existingRead)
            {
                return componentMessageManager.ReadExistingComponentsByType<TComponent>();
            }
            else if (pendingRead)
            {
                return componentMessageManager.ReadPendingComponentsByType<TComponent>();
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        protected ValueTuple<Guid, TComponent> ReadComponent<TComponent>() where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));
            if (existingRead && pendingRead)
            {
                return componentMessageManager.ReadFirstExistingOrPendingComponentByType<TComponent>();
            }
            else if (existingRead)
            {
                return componentMessageManager.ReadFirstExistingComponentByType<TComponent>();
            }
            else if (pendingRead)
            {
                return componentMessageManager.ReadFirstPendingComponentByType<TComponent>();
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        protected bool SomeComponent<TComponent>() where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));
            if (existingRead && pendingRead)
            {
                return componentMessageManager.SomeExistingOrPendingComponent<TComponent>();
            }
            else if (existingRead)
            {
                return componentMessageManager.SomeExistingComponent<TComponent>();
            }
            else if (pendingRead)
            {
                return componentMessageManager.SomePendingComponent<TComponent>();
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        protected IEnumerable<ValueTuple<Guid, TComponent>> GetComponents<TComponent>(Entity entity) where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));
            if (existingRead && pendingRead)
            {
                return componentMessageManager.ReadExistingAndPendingComponentsByEntityAndType<TComponent>(entity);
            }
            else if (existingRead)
            {
                return componentMessageManager.ReadExistingComponentsByEntityAndType<TComponent>(entity);
            }
            else if (pendingRead)
            {
                return componentMessageManager.ReadPendingComponentsByEntityAndType<TComponent>(entity);
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        protected ValueTuple<Guid, TComponent> GetComponent<TComponent>(Entity entity) where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));
            if (existingRead && pendingRead)
            {
                return componentMessageManager.ReadFirstExistingOrPendingComponentByEntityAndType<TComponent>(entity);
            }
            else if (existingRead)
            {
                return componentMessageManager.ReadFirstExistingComponentByEntityAndType<TComponent>(entity);
            }
            else if (pendingRead)
            {
                return componentMessageManager.ReadFirstPendingComponentByEntityAndType<TComponent>(entity);
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        protected bool HasComponent<TComponent>(Entity entity) where TComponent : struct, IComponent
        {
            var pendingRead = receiveTypes.Contains(typeof(PendingComponentMessage<TComponent>));
            var existingRead = receiveTypes.Contains(typeof(ComponentMessage<TComponent>));

            if (pendingRead && existingRead)
            {
                return componentMessageManager.HasExistingOrPendingComponent<TComponent>(entity);
            }
            else if (existingRead)
            {
                return componentMessageManager.HasExistingComponent<TComponent>(entity);
            }
            else if (pendingRead)
            {
                return componentMessageManager.HasPendingComponent<TComponent>(entity);
            }
            else
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }
        }

        internal void UpdateComponentInWorld<TComponent>(Guid componentID, TComponent newComponent) where TComponent : struct, IComponent
        {
            componentManager.AddUpdateComponentOperation(componentID, newComponent);
        }

        protected void UpdateComponent<TComponent>(Guid componentID, TComponent newComponentValue) where TComponent : struct, IComponent
        {
            if (!sendTypes.Contains(typeof(ComponentUpdateMessage<TComponent>)))
            {
                throw new IllegalUpdateException("Engine {0} tried to update undeclared Component {1}", GetType().Name, typeof(TComponent).Name);
            }

            ComponentUpdateMessage<TComponent> componentUpdateMessage;
            componentUpdateMessage.componentID = componentID;
            componentUpdateMessage.component = newComponentValue;
            SendMessage(componentUpdateMessage);
        }

        protected void SendMessage<TMessage>(TMessage message) where TMessage : struct, IMessage
        {
            if (!sendTypes.Contains(typeof(TMessage)))
            {
                throw new IllegalSendException("Engine {0} tried to send undeclared Message {1}", GetType().Name, typeof(TMessage).Name);
            }

            messageManager.AddMessage(message);
        }

        // unparameterized version to enable dynamic dispatch
        protected void SendMessage(IMessage message)
        {
            var type = message.GetType();

            if (!sendTypes.Contains(type) || !type.IsValueType)
            {
                throw new IllegalSendException("Engine {0} tried to send undeclared Message {1}", GetType().Name, type.Name);
            }

            messageManager.AddMessage(message);
        }

        internal void SendExistingComponentMessage<TComponent>(ComponentMessage<TComponent> message) where TComponent : struct, IComponent
        {
            componentMessageManager.AddExistingComponentMessage(message);
        }

        internal void SendPendingComponentMessage<TComponent>(PendingComponentMessage<TComponent> message) where TComponent : struct, IComponent
        {
            componentMessageManager.AddPendingComponentMessage(message);
        }

        protected IEnumerable<TMessage> ReadMessages<TMessage>() where TMessage : struct, IMessage
        {
            if (!receiveTypes.Contains(typeof(TMessage)))
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Message {1}", this.GetType().Name, typeof(TMessage).Name);
            }

            return messageManager.GetMessagesByType<TMessage>();
        }

        protected TMessage ReadMessage<TMessage>() where TMessage : struct, IMessage
        {
            return ReadMessages<TMessage>().Single();
        }

        protected bool SomeMessage<TMessage>() where TMessage : struct, IMessage
        {
            if (!receiveTypes.Contains(typeof(TMessage)))
            {
                throw new IllegalReadException("Engine {0} tried to read undeclared Message {1}", GetType().Name, typeof(TMessage).Name);
            }

            return ReadMessages<TMessage>().Any();
        }

        protected void Destroy(Guid entityID)
        {
            entityManager.MarkForDestroy(entityID);
        }

        protected void RemoveComponent(Guid componentID)
        {
            componentManager.MarkForRemoval(componentID);
        }
    }
}
